package iDatabase;
use strict;

################################################################
#
# iDatabase v1.0 (May 2001)
#
# Developed for Ikonboard.
# Author: Matthew Mecham <matt@ikonboard.com>
#
# Accessor methods to databases
#
# DBM: Interface to DBMS.
#
################################################################

BEGIN { @AnyDBM_File::ISA = qw(DB_File GDBM_File NDBM_File SDBM_File) }
use AnyDBM_File;

use vars qw($VERSION %DEBUG %GBL $INFO);
if ($^O ne 'MacOS' && ($^O ne 'MSWin32' || !Win32::IsWin95())) {
    use Fcntl qw(:DEFAULT :flock);
    $INFO->{'FLOCK'} = 1;
}

$VERSION = 1.0;
$INFO->{'do_lock'} = 1;

sub new {
    my $pkg = shift;

    # Collect the std. info, most is not needed for the DBM.

    my %args = (    "DATABASE" => "",
                    "IP"       => "",
                    "PORT"     => "",
                    "USERNAME" => "",
                    "PASSWORD" => "",
                    "DB_DIR"   => "",
                    @_,
               );
    my $obj = { 'error' => '', 'base_dir' => $args{'DB_DIR'} };
    bless $obj, $pkg;
    return $obj;
}


###############################################################################################
=pod

=head1 INDEX ROUTINES

Speeds up DBM searching by using the primary key as a foreign key
For example, Member Profiles: If we want to find a record based on the email address
Without an index, we'd have to parse every record to find a match.
Using an index, we can use the value queried as the key, and the value of the key/value
pair found as the foreign key.

It is the programmers responibility to keep the indexes up to date

=head2 EXAMPLE

We're looking for a member with an email address of matt\@ikonboard.com
    my $member = $db->query( TABLE => 'member_profiles',
                             MATCH => 'ONE',
                             INDEX => { KEY   => 'MEMBER_EMAIL',
                                        VALUE => 'matt\@ikonboard.com',
                                      }
                           );

When the query is called, iDatabase looks for an index called 'MEMBER_EMAIL.idx' in the member_profiles
directory. If it's found, it'll look for $DB{matt\@ikonboard.com}, the value is '77-98765432'.
Then, it'll open up the member_profiles.db and grab $DB{77-98765432} and parse the record and return it
as a hash reference.

=head2 CREATING INDEX

Continuing with our example,

    $db->create_index( TABLE       => 'member_profiles',
                       INDEX_KEY   => 'MEMBER_EMAIL',
                       FOREIGN_KEY => 'MEMBER_ID',
                     );

This creates a new .idx file, named after the INDEX_KEY.

=head 2 DROPPING INDEX

    $db->drop_index( TABLE     => 'member_profiles',
                     INDEX_KEY => 'MEMBER_EMAIL',
                   );

This simple unliks the .idx file

=head 2 REMOVING AN INDEX ENTRY

    $db->update_index( TABLE     => 'member_profiles',
                       INDEX_KEY => 'MEMBER_EMAIL',
                       R_KEY     => 'matt@ikonboard.com',
                       REMOVE    => 1
                     );

This removes the index entry for the key 'matt@ikonboard.com'

=head 2 UPDATING AN INDEX ENTRY

    $db->update_index( TABLE     => 'member_profiles',
                       INDEX_KEY => 'MEMBER_EMAIL',
                       R_KEY     => 'matt@ikonboard.com',
                       R_VALUE   => '77-123456789'
                     );
This updates the value of the R_KEY hash

=head 3 CAVETS

If you wish to update the KEY of the DBM entry, you'll have to REMOVE the index entry, then
update with the new R_KEY and R_VALUE's


=cut
###############################################################################################


sub create_index {
    my $obj = shift;
    my $IN = { TABLE       => "",
               DBID        => "",
               ID          => "", # DEPRECIATED IN DBM
               INDEX_KEY   => "",  #Column Name to use as an index key
               FOREIGN_KEY => "",  #Foreign Key to return
               @_,
             };

    $obj->load_cfg($IN->{'TABLE'});

    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';

    if ($obj->{'cur_method'} eq 'single') {
        $IN->{'ID'} =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';
    } else {
        $IN->{'ID'} = undef;
    }

    my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'INDEX_KEY'}$IN->{'ID'}.idx";

    $obj->lock_table($IN->{'TABLE'});

    tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";

    untie (%DB);

    $obj->release_lock($IN->{'TABLE'});
    
    return 1;
}


sub drop_index {
    my $obj = shift;
    my $IN = { TABLE       => "",
               DBID        => "",
               ID          => "", # DEPRECIATED IN DBM
               INDEX_KEY   => "",
               @_,
             };

    $obj->load_cfg($IN->{'TABLE'});

    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';

    if ($obj->{'cur_method'} eq 'single') {
        $IN->{'ID'} =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';
    } else {
        $IN->{'ID'} = undef;
    }

    my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'INDEX_KEY'}$IN->{'ID'}.idx";

    if (-e $file) {
        unlink $file;
        return 1;
    } elsif (-e $file.'.pag') {
        unlink "$file.pag";
        return 1;
    } else {
        return 0;
    }
}


sub update_index {
    my $obj = shift;
    my $IN = { TABLE       => "",
               DBID        => "",
               ID          => "", # DEPRECIATED IN DBM
               INDEX_KEY   => "",
               R_KEY       => "",
               R_VALUE     => "",
               REMOVE      => "",
               @_,
             };

    $obj->load_cfg($IN->{'TABLE'});

    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';

    if ($obj->{'cur_method'} eq 'single') {
        $IN->{'ID'} =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';
    } else {
        $IN->{'ID'} = undef;
    }

    my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'INDEX_KEY'}$IN->{'ID'}.idx";

    return unless ((-e $file) || (-e "$file.pag"));

    $obj->lock_table($IN->{'TABLE'});

    tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";
    if ($IN->{'REMOVE'}) {
        delete ( $DB{ $IN->{'R_KEY'} } );
    } else {
        $DB{ $IN->{'R_KEY'} } = $IN->{'R_VALUE'};
    }
    untie (%DB);

    $obj->release_lock($IN->{'TABLE'});
    
    return 1;
}


sub query_index {
    my $obj = shift;
    my $IN = { TABLE     => "",
               DBID      => "",
               ID        => "", # DEPRECIATED IN DBM
               INDEX_KEY => "",
               S_KEY     => "",
               CASE      => "",
               @_,
             };

    my $return;

    $obj->load_cfg($IN->{'TABLE'});

    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';

    if ($obj->{'cur_method'} eq 'single') {
        $IN->{'ID'} =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';
    } else {
        $IN->{'ID'} = undef;
    }

    my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'INDEX_KEY'}$IN->{'ID'}.idx";

    return unless ((-e $file) || (-e "$file.pag"));

    $obj->lock_table($IN->{'TABLE'});

    tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";

    if ($IN->{'CASE'}) {
        for my $k (keys %DB) {
            if ( lc($k) eq lc($IN->{'S_KEY'}) ) {
                $return = $DB{$k};
                last;
            }
        }
    } else {
        $return = $DB{ $IN->{'S_KEY'} };
    }

    untie (%DB);

    $obj->release_lock($IN->{'TABLE'});

    return $return;

}



#+---------------------------------------------------------------------------------------------
#| Create Table: Creates a table, all errors to $obj->{'error'}, returns 1 on success.
#+---------------------------------------------------------------------------------------------

sub create_table {
    my $obj = shift;
       $obj->{'error'} = undef;
    my $IN = {  
                "TABLE"   => "",
                "DBID"    => "",
                @_,
              };



    $obj->load_cfg($IN->{'TABLE'});

    if ($obj->{'cur_method'} eq 'single') {
        $IN->{'ID'} =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';
    } else {
        $IN->{'ID'} = undef;
    }


    # DBID is the table sub dir. (Table_name/dbid/)
    
    # iB3 has DBIDs for forum_posts and forum_polls  (forum_posts/f0 would be the table/dbid for the forum posts)
    # As a point of reference, the SQL version won't bother with that, it'll merely use "f0_forum_posts" as a table name

    unless (-e $obj->{'base_dir'} . $IN->{'TABLE'} ) {
        unless (mkdir ($obj->{'base_dir'} . $IN->{'TABLE'}, 0777) ) {
            $obj->{'error'} = "Could not create table $IN->{'TABLE'}! Please allow the Database dir to have write to permission";
            return;
        }
    }

    if ($IN->{'DBID'}) {
        if (-e $obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'}) {
            $obj->{'error'} = "Database Table ( $IN->{'TABLE'}/$IN->{'DBID'} ) already exists!";
            return;
        }
        unless (mkdir ($obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'}, 0777) ) {
            $obj->{'error'} = "Could not create table $IN->{'TABLE'}/$IN->{'DBID'}! Please allow the Database dir to have write to permission";
            return;
        }
        
        my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'TABLE'}$IN->{'ID'}.db";
        tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";
        untie (%DB);
        
    } else {
        if (-e $obj->{'base_dir'} . $IN->{'TABLE'}."/$IN->{'TABLE'}$IN->{'ID'}.db") {
            $obj->{'error'} = "Database Table ( $IN->{'TABLE'}/$IN->{'TABLE'}$IN->{'ID'}.db ) already exists!";
            return;
        }
        my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'TABLE'}$IN->{'ID'}.db";
        tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";
        untie (%DB);
    }      

    return 1;

}


sub drop_table {
    my $obj = shift;

    my %IN = ( TABLE    => "",
               DBID     => "",
               @_,
             );

    my $path = $IN{'DBID'} ? $obj->{'base_dir'}."$IN{'TABLE'}/$IN{'DBID'}" : $obj->{'base_dir'}.$IN{'TABLE'};

    opendir (DB, $path);
    my @files = grep { !/^\./ } readdir(DB);
    closedir(DB);

    for my $file (@files) {
        unless (-d $path.'/'.$file) {
            unlink $path.'/'.$file;
        }
    }

    if ($IN{'DBID'}) {
        rmdir $path;
    }

    return 1;
}


sub drop_tables {
    my $obj = shift;
    
    my %IN = ( TABLES => [],
               @_,
             );

    foreach my $table (@{$IN{'TABLES'}}) {
        $obj->drop_table( TABLE => $table->{'TABLE'},
                          DBID  => $table->{'DBID'},
                        );
    }
    return 1;
}


#+---------------------------------------------------------------------------------------------
#| SELECT: Returns a single record based on criteria
#+---------------------------------------------------------------------------------------------


sub select  {
    my $obj = shift;
       $obj->{'error'} = undef;
    my $IN = {  
                "TABLE"   => "",
                "COLUMNS" => [],
                "KEY"     => "",
                "DBID"    => "",
                "ID"      => "", # DEPRECIATED IN DBM
                @_
              };
    my ($data, $found);

    unless ( $obj->ping(TABLE=>$IN->{'TABLE'},DBID=>$IN->{'DBID'},ID=>$IN->{'ID'}) ) {
        $obj->{'error'} = 'Table Not Found!';
        return;
    }

    $obj->load_cfg($IN->{'TABLE'});


    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';

    if ($obj->{'cur_method'} eq 'single') {
        $IN->{'ID'} =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';
    } else {
        $IN->{'ID'} = undef;
    }

    my $cur_columns = scalar @{$IN->{'COLUMNS'}} > 0 ? \@{$IN->{'COLUMNS'}} : \@{$obj->{'col_name'}};

	my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'TABLE'}$IN->{'ID'}.db";
    return unless ((-e $file) || (-e "$file.pag"));

	$obj->lock_table($IN->{'TABLE'});

	tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";

	unless ( exists $DB{ $IN->{'KEY'} } ) {
    	$obj->{'error'} = "DB Error: Cannot locate the primary key whilst selecting $IN->{'KEY'} in table $IN->{'TABLE'}";
    	return;
  	}

	my $data = $obj->decode_record( $DB{ $IN->{'KEY'} });

	untie (%DB);

	$obj->release_lock($IN->{'TABLE'});
    
    my $return = {};
    if (scalar @{$IN->{'COLUMNS'}} != $obj->{'total_cols'}) {
        for (@{$cur_columns}) {
            $return->{$_} = $data->{$_};
        }
    } else {
        $return = $data;
    }
    
    return $return ? $return : 0;
}


#+---------------------------------------------------------------------------------------------
#| QUERY: Returns records based on criteria(s)
#+---------------------------------------------------------------------------------------------


sub query  {
    my $obj = shift;
       $obj->{'error'} = undef;
    my $IN = {  
                "TABLE"     => "",
                "COLUMNS"   => [],
                "SORT_KEY"  => "",
                "SORT_BY"   => "",
                "WHERE"     => "",
                "MATCH"     => "",
                "RANGE"     => "",
                "DBID"      => "",
                "ID"        => "", #DEPRECIATED IN DBM
                "COUNT"     => "",
                "POP"       => "",
                "INDEX"     => "",
                @_,
              };

    if ($IN->{'INDEX'}) {

        my $data = $obj->query_index( TABLE     => $IN->{'TABLE'},
                                      DBID      => $IN->{'DBID'},
                                      ID        => $IN->{'ID'},
                                      INDEX_KEY => $IN->{'INDEX'}->{'KEY'},
                                      S_KEY     => $IN->{'INDEX'}->{'VALUE'},
                                      CASE      => $IN->{'INDEX'}->{'CASE'}
                                    );

        return {} unless defined $data;

        my $return = $obj->select( TABLE     => $IN->{'TABLE'},
                                   DBID      => $IN->{'DBID'},
                                   ID        => $IN->{'ID'},
                                   KEY       => $data
                                 );

        return $return;
    }


    my ($data, $found, $statement_string, $returned_records, @return, @found, $return_single, @popped);

    unless ( $obj->ping(TABLE=>$IN->{'TABLE'},DBID=>$IN->{'DBID'}) ) {
        $obj->{'error'} = 'Table Not Found!';
        return;
    }

    $obj->load_cfg($IN->{'TABLE'});

    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';

    if ($obj->{'cur_method'} eq 'single') {
        $IN->{'ID'} =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';
    } else {
        $IN->{'ID'} = undef;
    }

    my $cur_columns = scalar @{$IN->{'COLUMNS'}} > 0 ? \@{$IN->{'COLUMNS'}} : \@{$obj->{'col_name'}};

    #Build the search query
    
    if ($IN->{'WHERE'}) {
        $statement_string = "(".$IN->{'WHERE'}.")";
        for my $i (0 .. $obj->{'total_cols'}) {
            $statement_string =~ s!([\A\s\b\(])($obj->{'col_name'}->[$i])([\Z\b\s\)])!$1\$data->{'$2'}$3!g;
        }
    } else {
		$statement_string = qq~(\$data->{ \$obj->{'cur_p_key'} } ne '')~;
	}

    if ($IN->{'POP'}) {
        $IN->{'POP'} =~ s{\A(\({0,1})(\S+) (<|>|=|==|eq|ne|>=|=<|=~|!~) ([\(/"'\s])([\S\s\w\d]+)([\)/i"'\s])(\){0,1})\Z}{$1(\$data->{'$2'} $3 $4$5$6$7)}isg;
        my $code = qq~ sub POP {  my (\$obj, \$data) = \@_; return 1 if $IN->{'POP'}; } ~;
        eval $code; die $@ if $@;
    }

    #die "$statement_string";
    #die "$IN->{'POP'}";
    my $found_records = {};

    my $eval = qq~sub Check {  my (\$obj, \$data) = \@_; return 1 if $statement_string; }~;
    
    eval $eval; die "Eval Error while parsing:: $statement_string" . $@ if $@;

	my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'TABLE'}$IN->{'ID'}.db";
	return my $empty_array = $IN->{'MATCH'} eq 'ONE' ? {} : [] unless ((-e $file) || (-e "$file.pag"));

	$obj->lock_table($IN->{'TABLE'});

	tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";

	for (keys %DB) {
		my $data = $obj->decode_record($DB{$_});
		if ($obj->Check($data)) {
    		$found_records->{$data->{$obj->{'cur_p_key'}}} = $data;
    		++$returned_records;
    		if ($returned_records == 1 and $IN->{'MATCH'} eq 'ONE') {
        		$found = $data;
       			last;
			}
		}
	}

	untie (%DB);

	$obj->release_lock($IN->{'TABLE'});

    if ($IN->{'MATCH'} ne 'ONE') {

        goto RANGE if $IN->{'COUNT'};

        if ($obj->{'all_cols'}->{ $IN->{'SORT_KEY'} }[1] eq 'string') {
            if ($IN->{'SORT_BY'} eq 'Z-A') {
                for (sort { $found_records->{$b}->{ $IN->{'SORT_KEY'} } cmp $found_records->{$a}->{ $IN->{'SORT_KEY'} } } keys %{$found_records}) {
                    if ($IN->{'POP'} and $obj->POP($found_records->{$_})) { push @popped, $found_records->{$_} } else { push @return, $found_records->{$_}; }
                }
            } else {
                for (sort { $found_records->{$a}->{ $IN->{'SORT_KEY'} } cmp $found_records->{$b}->{ $IN->{'SORT_KEY'} } } keys %{$found_records}) {
                    if ($IN->{'POP'} and $obj->POP($found_records->{$_})) { push @popped, $found_records->{$_} } else { push @return, $found_records->{$_}; }
                }
            }
        } else {
            if ($IN->{'SORT_BY'} eq 'Z-A') {
                for (sort { $found_records->{$b}->{ $IN->{'SORT_KEY'} } <=> $found_records->{$a}->{ $IN->{'SORT_KEY'} } } keys %{$found_records}) {
                    if ($IN->{'POP'} and $obj->POP($found_records->{$_})) { push @popped, $found_records->{$_} } else { push @return, $found_records->{$_}; }
                }
            } else {
			    for (sort { $found_records->{$a}->{ $IN->{'SORT_KEY'} } <=> $found_records->{$b}->{ $IN->{'SORT_KEY'} } } keys %{$found_records}) {
                    if ($IN->{'POP'} and $obj->POP($found_records->{$_})) { push @popped, $found_records->{$_} } else { push @return, $found_records->{$_}; }
                }
            }
        }
        
        if (@popped > 0) {
            @return = (@popped, @return);
        }

        RANGE:

        if ($IN->{'RANGE'}) {
            my ($from, $to) = split " to ", $IN->{'RANGE'};
            $from = 0 unless $from;
            $to   = $#return if ((!$to) or ($to > $#return));
            @return = @return[$from .. $to];
        }

        @return = () unless scalar @return > 0;
        if ($IN->{'COUNT'}) {
            return $returned_records;
        } else {
            return \@return;
        }
    }

    return $returned_records == 1 ? $found : {};
}


#+---------------------------------------------------------------------------------------------
#| INSERT: Inserts a record into the table
#+---------------------------------------------------------------------------------------------


sub insert {
    my $obj = shift;
       $obj->{'error'} = undef;
    my $IN = {  
                "TABLE"    => "",
                "VALUES"   => "",
                "DBID"     => "",
                "ID"       => "", #DEPRECIATED IN DBM
                @_
              };  


    unless ( $obj->ping(TABLE=>$IN->{'TABLE'},DBID=>$IN->{'DBID'}) ) {
        $obj->{'error'} = 'Table Not Found!';
        return;
    }

    $obj->load_cfg($IN->{'TABLE'});

    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';

    if ($obj->{'cur_method'} eq 'single') {
        $IN->{'ID'} =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';
    } else {
        $IN->{'ID'} = undef;
    }

    my $counter;
    
    if ($obj->{'all_cols'}->{ $obj->{'cur_p_key'} }[1] eq 'update') {
        open (CNT, $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'TABLE'}$IN->{'ID'}.cnt.db");
        flock (CNT, LOCK_SH) if $INFO->{'FLOCK'};
        $counter = <CNT> || 0;
        close (CNT);

        ++$counter;

        $IN->{'VALUES'}->{ $obj->{'cur_p_key'} } = $counter;
    } 


	my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'TABLE'}$IN->{'ID'}.db";
	#return unless (-e $file);

	my $entry = $obj->encode_record($IN->{'VALUES'});

	$obj->lock_table($IN->{'TABLE'});

	tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";

    $DB{ $IN->{'VALUES'}->{ $obj->{'cur_p_key'} } } = $entry;

	untie (%DB);
        
	$obj->release_lock($IN->{'TABLE'});

  
    if ($obj->{'all_cols'}->{ $obj->{'cur_p_key'} }[1] eq 'update') {
        open (CNT, ">$obj->{'base_dir'}" . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'TABLE'}$IN->{'ID'}.cnt.db") or die $!;
        flock (CNT, LOCK_EX) or die $! if $INFO->{'FLOCK'};
        print CNT $counter;
        close (CNT) or die $!;

        return $counter;
    }
}



#+---------------------------------------------------------------------------------------------
#| DELETE: Deletes a record(s) based on the Primary Key
#+---------------------------------------------------------------------------------------------

sub delete  {
    my $obj = shift;
       $obj->{'error'} = undef;
    my $IN = {  
                "TABLE"   => "",
                "KEY"     => "",
                "DBID"    => "",
                "ID"      => "", #DEPRECIATED IN DBM
                "WHERE"   => "",
                @_
              };

    my @keys;

    $obj->load_cfg($IN->{'TABLE'});

    if ($IN->{'KEY'} eq '') {
        return $obj->{error} = 'no keys to delete!' unless $IN->{'WHERE'} ne '';
    }

    unless ( $obj->ping(TABLE=>$IN->{'TABLE'},DBID=>$IN->{'DBID'}) ) {
        $obj->{'error'} = 'Table Not Found!';
        return;
    }

    if ($IN->{'WHERE'} ne '') {
        my $ids = $obj->query( TABLE   => $IN->{'TABLE'},
                               ID      => $IN->{'ID'},
                               COLUMNS => [$obj->{'cur_p_key'}],
                               WHERE   => $IN->{'WHERE'},
                               DBID    => $IN->{'DBID'},
                              );
        return unless scalar @{$ids} > 0;
        for (@{$ids}) {
            push @keys, $_->{ $obj->{'cur_p_key'} };
        }
    } else {
		if (ref($IN->{'KEY'}) ne 'ARRAY') {
			push @keys, $IN->{'KEY'};
		} else {
			for (@{$IN->{'KEY'}}) {
				push @keys, $_;
			}
		}
	}

    if ($obj->{'cur_method'} eq 'single') {
        $IN->{'ID'} =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';
    } else {
        $IN->{'ID'} = undef;
    }

    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';

	my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'TABLE'}$IN->{'ID'}.db";
    return unless ((-e $file) || (-e "$file.pag"));

	$obj->lock_table($IN->{'TABLE'});

	tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";

	for my $key (@keys) {
    	delete $DB{ $key };
	}

	untie (%DB);
        
	$obj->release_lock($IN->{'TABLE'});

    return 1;
}

#+---------------------------------------------------------------------------------------------
#| UPDATE: Updates a record(s) based on the Primary Key
#+---------------------------------------------------------------------------------------------

sub update  {
    my $obj = shift;
       $obj->{'error'} = undef;
    my $IN = {  
                "TABLE"   => "",
                "VALUES"  => "",
                "KEY"     => "",
                "WHERE"   => "",
                "DBID"    => "",
                "ID"      => "", #DEPRECIATED IN DBM
                @_
              };

    my ($data, $found);

    unless ( $obj->ping(TABLE=>$IN->{'TABLE'},DBID=>$IN->{'DBID'}) ) {
        $obj->{'error'} = 'Table Not Found!';
        return;
    }

    my @keys;

    $obj->load_cfg($IN->{'TABLE'});

    #$IN->{'VALUES'}->{ $obj->{'cur_p_key'} } = $IN->{'KEY'};

    if ($IN->{'WHERE'} ne '') {
        my $ids = $obj->query( TABLE   => $IN->{'TABLE'},
                               ID      => $IN->{'ID'},
                               COLUMNS => [$obj->{'cur_p_key'}],
                               WHERE   => $IN->{'WHERE'},
                               DBID    => $IN->{'DBID'},
                              );
        return unless scalar @{$ids} > 0;
        @keys = map {   $_->{ $obj->{'cur_p_key'} }   } @{$ids};
    } else {
		if (ref($IN->{'KEY'}) ne 'ARRAY') {
			push @keys, $IN->{'KEY'};
		} else {
			for (@{$IN->{'KEY'}}) {
				push @keys, $_;
			}
		}
	}
    
    return unless @keys > 0;

    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';

    if ($obj->{'cur_method'} eq 'single') {
        $IN->{'ID'} =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';
    } else {
        $IN->{'ID'} = undef;
    }

	my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'TABLE'}$IN->{'ID'}.db";
    return unless ((-e $file) || (-e "$file.pag"));

	$obj->lock_table($IN->{'TABLE'});

	tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";

    for my $key (@keys) {

	    my $data = $obj->decode_record( $DB{ $key });

	    for my $k (keys %{$IN->{'VALUES'}}) {
    	    $data->{$k} = $IN->{'VALUES'}->{$k};
	    }
	    my $entry = $obj->encode_record($data);

	    $DB{ $key } = $entry;
    }

	untie (%DB);
        
	$obj->release_lock($IN->{'TABLE'});

    return 1;
}


sub count {
    my $obj = shift;
    my $IN = {  
                "TABLE"    => "",
                "DBID"     => "",
                "ID"       => "",
                @_,
              };

    my $count = 0;

    $obj->load_cfg($IN->{'TABLE'});

    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';

    if ($obj->{'cur_method'} eq 'single') {
        $IN->{'ID'} =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';
    } else {
        $IN->{'ID'} = undef;
    }

    if ($IN->{'ID'}) {
	    my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'TABLE'}$IN->{'ID'}.db";
        return 0 unless ((-e $file) || (-e "$file.pag"));
	    $obj->lock_table($IN->{'TABLE'});
	    tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";
        for (keys %DB) {
            $count++;
        }
	    untie (%DB);
	    $obj->release_lock($IN->{'TABLE'});
        return $count;

    } else {
        opendir CNT, $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}" or die "Cannot open directory for reading";
        my @files = grep { /$IN->{'TABLE'}\-([\d\w\-\_])\.db/ } readdir CNT;
        closedir CNT;
        if (scalar @files > 1) {
            $IN->{'DBID'} =~ s!/+\Z!!;
            for my $f (@files) {
                $f =~ s/$IN->{'TABLE'}\-([\d\w\-\_])\.db/$1/;
                $count += $obj->count( TABLE    => $IN->{'TABLE'},
                                       DBID     => $IN->{'DBID'},
                                       ID       => $f,
                                     );
            }
            return $count;
        } else {
	        my $file = $obj->{'base_dir'} . "$IN->{'TABLE'}/$IN->{'DBID'}$IN->{'TABLE'}.db";
            return 0 unless ((-e $file) || (-e "$file.pag"));
	        $obj->lock_table($IN->{'TABLE'});
	        tie (my %DB, $AnyDBM_File::ISA[0], $file, O_RDWR|O_CREAT, 0777) || die "Can't open file ($file) for reading. $!";
            for (keys %DB) {
                $count++;
            }
	        untie (%DB);
	        $obj->release_lock($IN->{'TABLE'});
            return $count;
        }
    }
}
        
        


#+---------------------------------------------------------------------------------------------
#| ROLLBACK: Undoes the last database change.
#+---------------------------------------------------------------------------------------------

sub rollback  {
    my $obj = shift;
       $obj->{'error'} = undef;
    my $IN = {  
                "TABLE"   => "",
                "DBID"    => "",
                "ID"      => "",
                @_
              };

    unless ( $obj->ping(TABLE=>$IN->{'TABLE'},DBID=>$IN->{'DBID'},ID=>$IN->{'ID'}) ) {
        $obj->{'error'} = 'Table Not Found!';
        return;
    }

    unless (-e $obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'}.'file'.$IN->{'ID'}.'.bak.cgi') {
        $obj->{'error'} = 'Back-up not found!';
        return;
    }        

    my $old   = $obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'}.'file'.$IN->{'ID'}.'.bak.cgi';
    my $new   = $obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'}.'file'.$IN->{'ID'}.'.cgi'; 

    rename($old, $new);

    return 1;

}

#+---------------------------------------------------------------------------------------------
#| PING: Test to see if the table exists.
#+---------------------------------------------------------------------------------------------

sub ping {
	my $obj = shift;
    my $IN = {  
                "TABLE"   => "",
                "DBID"    => "",
                "ID"      => "",
                @_
              };
	
    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';
    $IN->{'ID'}   =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';

	return 1 if (-e $obj->{'base_dir'} . $IN->{'TABLE'});
}




sub export {
    my $obj = shift;

    my $IN = {  
                "TABLE"   => "",
                "DBID"    => "",
                "ID"      => "",
                @_
              };
	
    $IN->{'DBID'} =  $IN->{'DBID'} ne '' ? $IN->{'DBID'}.'/' : '';
    $IN->{'ID'}   =  $IN->{'ID'}   ne '' ? '-'.$IN->{'ID'}   : '';

    $obj->load_cfg($IN->{'TABLE'});

    my $export;

    if ($obj->{'cur_method'} eq 'single') {
        local $/ = undef;
        open EXP, $obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'}.'file'.$IN->{'ID'}.'.cgi' or $obj->{'error'} = "Could not access that table ($IN->{'TABLE'})";
        $export = <EXP>;
        close EXP;
    } else {

        # Push all entries into one table.
        # Remove Trailing slash if a DBID is defined.

        $IN->{'DBID'} =~ s!/+\Z!!;

        opendir DIR, $obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'} or die "Cannot Open the Directory!";
        my @files =  grep  {   !/^\./   } readdir DIR;
        closedir DIR;
        my @all_files;
        for my $file (@files) {
            next if $file =~ /\.bak\.cgi\Z/;
            next if $file =~ /\.cnt\.cgi\Z/;
            $IN->{'ID'} = $file;
            $IN->{'ID'} =~ s!file(.+?)\.cgi!$1!g;
            my $data = $obj->my_load_table_entry($IN);
            $data    = $obj->encode_record($data);
            push @all_files, "|*|$file|*|".$data;
        }
        $export = join "\n", @all_files;
    }

    if (-e $obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'}.'file'.$IN->{'ID'}.'.cnt.cgi') {
        open CNT, $obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'}.'file'.$IN->{'ID'}.'.cnt.cgi';
        my $cnt = <CNT>;
        close CNT;
        chomp $cnt;

        $export .= "#".$cnt;
    }

    # return table

    return $export;
}



sub lock_table {
    my ($obj, $table_name) = @_;
    my $EndTime = 30 + time;
    my $file = $table_name . '.lck';  
    while ((-e $file) && (time < $EndTime)) {
        sleep(1);
    }

    chmod (0777, $obj->{'base_dir'}.'Temp');
    open(LOCK, ">".$obj->{'base_dir'}.'Temp/'.$file) or $obj->{'error'} = "Can't open the lock file ($!)";
    chmod (0777, $obj->{'base_dir'}.'Temp/'.$file);
}


sub release_lock {
    my ($obj, $table_name) = @_;
    my $file = $table_name . '.lck';
    close(LOCK);
    unlink($obj->{'base_dir'}.'Temp/'.$file);
}







sub DESTROY { return undef; }

#+---------------------------------------------------------------------------------------------
#| DECODE_RECORD: Decode a record entry
#+---------------------------------------------------------------------------------------------

sub decode_record {
    my ($obj, $record) = @_;
    my $return = {};
    chomp $record;
    my @Tmp = split (/\Q|^|\E/, $record);
    for my $i (0 .. $obj->{'total_cols'}) {
        $Tmp[$i] =~ s!^\s+!!g;
        $Tmp[$i] =~ s!\s+$!!g;
        $Tmp[$i] =~ s!\$!&#36!g;
        $Tmp[$i] =~ s!\\n!\n!g;
        $return->{ $obj->{'col_name'}->[$i] } = $Tmp[$i];
        
    }

    return $return;
}
 
#+---------------------------------------------------------------------------------------------
#| ENCODE_RECORD: Encode a record entry
#+---------------------------------------------------------------------------------------------


sub encode_record {
    my ($obj, $values) = @_;
    my ($return, $cnt);
    for my $i (0 .. $obj->{'total_cols'}) {
        $values->{$obj->{'col_name'}->[$i]}  =~ s!^\s+!!g;
        $values->{$obj->{'col_name'}->[$i]}  =~ s!\s+$!!g;
        $values->{$obj->{'col_name'}->[$i]}  =~ s!\r!!g;
        $values->{$obj->{'col_name'}->[$i]}  =~ s!\n!\\n!g;
        $return .= $values->{$obj->{'col_name'}->[$i]}."|^|";
    }
  $return =~ s!\Q|^|\E\Z!!g;
  return $return;
}


#+---------------------------------------------------------------------------------------------
#| LOAD_CFG: Load and set up the configuration file.
#+---------------------------------------------------------------------------------------------

sub load_cfg ($$) {
    my ($obj, $cfg) = @_;
    die "Invalid Number of ARGVS" unless @_ == 2;
    do $obj->{'base_dir'}.'config/'.$cfg.'.cfg' or die "Cannot open file : $cfg.cfg ($!)";

    $obj->{'cur_table'}  = $IMPORT::STRING->{'TABLE'};
    $obj->{'cur_p_key'}  = $IMPORT::STRING->{'P_KEY'};
    $obj->{'cur_method'} = $IMPORT::STRING->{'MTD'};
    $obj->{'cur_update'} = $IMPORT::STRING->{'UPDATE'} || 'bottom';
    $obj->{'all_cols'}   = $IMPORT::COLS;
    $obj->{'total_cols'} = 0;
    $obj->{'col_name'}   = [];
    foreach (sort { $obj->{'all_cols'}->{$a}[0] <=> $obj->{'all_cols'}->{$b}[0] } keys %{$obj->{'all_cols'}} ) {
        push @{$obj->{'col_name'}}, $_;
        ++$obj->{'total_cols'};
    }
    $obj->{'total_cols'}--;
}


#+---------------------------------------------------------------------------------------------
#| Single Entry Data Types: File require's, private use only
#+---------------------------------------------------------------------------------------------


sub my_load_table_entry {
    my ($obj, $IN) = @_;
    return unless (-e $obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'}.'file'.$IN->{'ID'}.'.cgi');
    do $obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'}.'file'.$IN->{'ID'}.'.cgi' or die $!;
    return eval { $IN->{'TABLE'}->new() };
}


sub my_update_table_entry {
    my ($obj, $IN) = @_;
    my $data = $obj->my_load_table_entry($IN);
    for (keys %{ $IN->{'VALUES'} }) {
        $data->{$_} = $IN->{'VALUES'}->{$_};
    }
    $IN->{'VALUES'} = $data;
    $obj->_do_table_entry($IN);
}


sub my_insert_table_entry {
    my ($obj, $IN) = @_;
    $obj->_do_table_entry($IN);
}


sub my_delete_table_entry {
    my ($obj, $IN) = @_;
    unlink ($obj->{'base_dir'} . $IN->{'TABLE'}.'/'.$IN->{'DBID'}.'file'.$IN->{'ID'}.'.cgi');
    return 1;
}


sub _do_table_entry {
    my ($obj, $IN) = @_;
    my $data = $IN->{'VALUES'};
    open  (FH, ">$obj->{'base_dir'}" . $IN->{'TABLE'}.'/'.$IN->{'DBID'}.'file'.$IN->{'ID'}.'.cgi') or die $!;
    flock (FH, LOCK_EX) or die $! if $INFO->{'FLOCK'};

    print FH qq|package $IN->{'TABLE'};
                sub new {
                my \$pkg = shift;
                my \$obj = {
               |;
    
    foreach (keys %{$data}) {
        my $space = " " x (20 - (length($_)));
        if (ref($data->{$_}) eq 'ARRAY') {
            my $line = qq|        '$_' $space => [|;
            for my $i (@{ $data->{$_} }) { $line .= qq| "$i",|; }
            $line .= "]";
            $line =~ s!,\]!\]!;
            print FH $line.",\n";
        } else {
        print FH qq|        '$_' $space => q~$data->{$_}~,\n|;
        }
    }
    print FH qq|};
                bless \$obj, \$pkg;
                return \$obj;
                }
                1;|;
    close (FH) or die $!;
    return 1;
 }


#+---------------------------------------------------------------------------------------------
1;       # End of Module
#+---------------------------------------------------------------------------------------------

__END__


=pod

=head1 NAME

iDatabase: Universal Interface to database access

=head1 SYNOPSIS

    use iDatabase::CSV;

    my $db = iDatabase->new();

    my $key = $record_id;

    my $array_fef = $db->select( TABLE   => 'accounts',
                                 COLUMNS => ['ID', 'PRICE', 'COLOR'],
                                 KEY     => $key
                               );

    for $rv (@{$array_ref}) {
        print "Id Numer: $rv->{'ID'}, Price: $rv->{'PRICE'}, Color: $rv->{'COLOR'}\n";
    }

    $db->update( TABLE  => 'accounts',
                 KEY    => $key,
                 VALUES => { PRICE => '3.00' }
               );

=head1 DESCRIPTION

iDatabase is an independant interface for database access. From text files, to DBM hashes to SQL RDMS

It allows a script author to change database types on a whim.
<More later>

=cut






